import abc
from typing import Optional

import Levenshtein
import numpy as np
import pandas as pd
from jarowinkler import jarowinkler_similarity

from upgini.autofe.operator import PandasOperator, VectorizableMixin


class Min(PandasOperator):
    name: str = "min"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return np.minimum(left, right)


class Max(PandasOperator):
    name: str = "max"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return np.maximum(left, right)


class Add(PandasOperator, VectorizableMixin):
    name: str = "+"
    alias: str = "add"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True
    is_vectorizable: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left + right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.add(d2, axis=0)


class Subtract(PandasOperator, VectorizableMixin):
    name: str = "-"
    alias: str = "sub"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True
    is_vectorizable: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left - right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.sub(d2, axis=0)


class Multiply(PandasOperator, VectorizableMixin):
    name: str = "*"
    alias: str = "mul"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True
    is_vectorizable: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left * right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.mul(d2, axis=0)


class Divide(PandasOperator, VectorizableMixin):
    name: str = "/"
    alias: str = "div"
    is_binary: bool = True
    has_symmetry_importance: bool = True
    is_vectorizable: bool = True
    output_type: Optional[str] = "float"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left / right.replace(0, np.nan)

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.div(d2.replace(0, np.nan), axis=0)


class Combine(PandasOperator):
    name: str = "Combine"
    is_binary: bool = True
    has_symmetry_importance: bool = True
    output_type: Optional[str] = "object"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        temp = left.astype(str) + "_" + right.astype(str)
        temp[left.isna() | right.isna()] = np.nan
        return pd.Series(temp, index=left.index)


class CombineThenFreq(PandasOperator):
    name: str = "CombineThenFreq"
    is_binary: bool = True
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True
    output_type: Optional[str] = "float"
    is_distribution_dependent: bool = True
    input_type: Optional[str] = "discrete"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        temp = left.astype(str) + "_" + right.astype(str)
        temp[left.isna() | right.isna()] = np.nan
        value_counts = temp.value_counts(normalize=True)
        self._loc(temp, value_counts)


class Distance(PandasOperator):
    name: str = "dist"
    is_binary: bool = True
    output_type: Optional[str] = "float"
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return pd.Series(
            1 - self.__dot(left, right) / (self.__norm(left) * self.__norm(right)), index=left.index
        ).astype(np.float64)

    # row-wise dot product, handling None values
    def __dot(self, left: pd.Series, right: pd.Series) -> pd.Series:
        left = left.apply(lambda x: np.array(x).astype(np.float64))
        right = right.apply(lambda x: np.array(x).astype(np.float64))
        res = (left.dropna() * right.dropna()).apply(np.sum)
        res = res.reindex(left.index.union(right.index))
        return res

    # Calculate the norm of a vector, handling None values
    def __norm(self, vector: pd.Series) -> pd.Series:
        vector = vector.fillna(np.nan)
        return np.sqrt(self.__dot(vector, vector))


# Left for backward compatibility
class Sim(Distance):
    name: str = "sim"
    is_binary: bool = True
    output_type: Optional[str] = "float"
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return 1 - super().calculate_binary(left, right)


class StringSim(PandasOperator, abc.ABC):
    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        sims = []
        for i in left.index:
            left_i = self._prepare_value(left.get(i))
            right_i = self._prepare_value(right.get(i))
            if left_i is not None and right_i is not None:
                sims.append(self._similarity(left_i, right_i))
            else:
                sims.append(None)

        return pd.Series(sims, index=left.index)

    @abc.abstractmethod
    def _prepare_value(self, value: Optional[str]) -> Optional[str]:
        pass

    @abc.abstractmethod
    def _similarity(self, left: str, right: str) -> float:
        pass


class JaroWinklerSim1(StringSim):
    name: str = "sim_jw1"
    is_binary: bool = True
    input_type: Optional[str] = "string"
    output_type: Optional[str] = "float"
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def _prepare_value(self, value: Optional[str]) -> Optional[str]:
        return value if value is not None and len(value) > 0 else None

    def _similarity(self, left: str, right: str) -> float:
        return jarowinkler_similarity(left, right)


class JaroWinklerSim2(StringSim):
    name: str = "sim_jw2"
    is_binary: bool = True
    input_type: Optional[str] = "string"
    output_type: Optional[str] = "float"
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def _prepare_value(self, value: Optional[str]) -> Optional[str]:
        return value[::-1] if value is not None and len(value) > 0 else None

    def _similarity(self, left: str, right: str) -> float:
        return jarowinkler_similarity(left, right)


class LevenshteinSim(StringSim):
    name: str = "sim_lv"
    is_binary: bool = True
    input_type: Optional[str] = "string"
    output_type: Optional[str] = "float"
    is_symmetrical: bool = True
    has_symmetry_importance: bool = True

    def _prepare_value(self, value: Optional[str]) -> Optional[str]:
        return value if value is not None and len(value) > 0 else None

    def _similarity(self, left: str, right: str) -> float:
        return 1 - Levenshtein.distance(left, right) / max(len(left), len(right))



from typing import List

import pandas as pd


class BaseSearchKeyDetector:
    def _is_search_key_by_name(self, column_name: str) -> bool:
        raise NotImplementedError

    def _is_search_key_by_values(self, column: pd.Series) -> bool:
        raise NotImplementedError

    def _get_search_keys_by_name(self, column_names: List[str]) -> List[str]:
        return [
            column_name
            for column_name in column_names
            if self._is_search_key_by_name(column_name)
        ]

    def get_search_key_columns(self, df: pd.DataFrame, existing_search_keys: List[str]) -> List[str]:
        other_columns = [col for col in df.columns if col not in existing_search_keys]
        columns_by_names = self._get_search_keys_by_name(other_columns)
        columns_by_values = []
        for column_name in other_columns:
            if self._is_search_key_by_values(df[column_name]):
                columns_by_values.append(column_name)

        both = [col for col in columns_by_names if col in columns_by_values]
        only_values = [col for col in columns_by_values if col not in columns_by_names]
        only_names = [col for col in columns_by_names if col not in columns_by_values]
        return both + only_values + only_names



import pandas as pd
from typing import Dict, Optional

from upgini.autofe.operator import ParametrizedOperator
from upgini.autofe.timeseries.base import TimeSeriesBase
from upgini.autofe.utils import pydantic_validator

# Roll aggregation functions
roll_aggregations = {
    "norm_mean": lambda x: x[-1] / x.mean(),
    "q25": lambda x: x.quantile(0.25),
    "q75": lambda x: x.quantile(0.75),
    "iqr": lambda x: x.quantile(0.75) - x.quantile(0.25),
}


class Roll(TimeSeriesBase, ParametrizedOperator):
    aggregation: str
    window_size: int = 1
    window_unit: str = "D"

    @pydantic_validator("window_unit")
    def validate_window_unit(cls, v: str) -> str:
        try:
            pd.tseries.frequencies.to_offset(v)
            return v
        except ValueError:
            raise ValueError(
                f"Invalid window_unit: {v}. Must be a valid pandas frequency string (e.g. 'D', 'H', 'T', etc)"
            )

    def to_formula(self) -> str:
        # First add window size and unit, then add aggregation, then add offset
        base_formula = f"roll_{self.window_size}{self.window_unit}"
        formula_with_agg = f"{base_formula}_{self.aggregation}"
        return self._add_offset_to_formula(formula_with_agg)

    @classmethod
    def from_formula(cls, formula: str) -> Optional["Roll"]:
        import re

        # Base regex for Roll class (with aggregation)
        base_regex = r"roll_(\d+)([a-zA-Z])_(\w+)"

        # Parse offset first - this removes the offset part if present
        offset_params, remaining_formula = cls._parse_offset_from_formula(formula, base_regex)

        if remaining_formula is None:
            return None

        # Parse the window part and aggregation
        match = re.match(f"^{base_regex}$", remaining_formula)

        if not match:
            return None

        window_size = int(match.group(1))
        window_unit = match.group(2)
        aggregation = match.group(3)

        # Create instance with appropriate parameters
        params = {
            "window_size": window_size,
            "window_unit": window_unit,
            "aggregation": aggregation,
        }

        if offset_params:
            params.update(offset_params)

        return cls(**params)

    def get_params(self) -> Dict[str, Optional[str]]:
        res = super().get_params()
        res.update(
            {
                "window_size": self.window_size,
                "window_unit": self.window_unit,
                "aggregation": self.aggregation,
            }
        )
        return res

    def _aggregate(self, ts: pd.DataFrame) -> pd.DataFrame:
        return ts.rolling(f"{self.window_size}{self.window_unit}", min_periods=1).agg(
            roll_aggregations.get(self.aggregation, self.aggregation)
        )



import logging
from logging import Logger
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_integer_dtype, is_object_dtype, is_string_dtype

from upgini.resource_bundle import bundle


class FeaturesValidator:
    def __init__(self, logger: Optional[Logger] = None):
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger()
            self.logger.setLevel("FATAL")

    def validate(
        self,
        df: pd.DataFrame,
        features: List[str],
        features_for_generate: Optional[List[str]] = None,
        columns_renaming: Optional[Dict[str, str]] = None,
    ) -> Tuple[List[str], List[str]]:
        one_hot_encoded_features = []
        empty_or_constant_features = []
        high_cardinality_features = []
        warnings = []

        for f in features:
            column = df[f]
            if is_object_dtype(column):
                column = column.astype("string")
            value_counts = column.value_counts(dropna=False, normalize=True)
            most_frequent_percent = value_counts.iloc[0]

            if len(value_counts) == 1:
                empty_or_constant_features.append(f)
            elif most_frequent_percent >= 0.99:
                if self.is_one_hot_encoded(column):
                    one_hot_encoded_features.append(f)
                else:
                    empty_or_constant_features.append(f)

        columns_renaming = columns_renaming or {}

        if one_hot_encoded_features and len(one_hot_encoded_features) > 1:
            msg = bundle.get("one_hot_encoded_features").format(
                [columns_renaming.get(f, f) for f in one_hot_encoded_features]
            )
            warnings.append(msg)

        if empty_or_constant_features:
            msg = bundle.get("empty_or_contant_features").format(
                [columns_renaming.get(f, f) for f in empty_or_constant_features]
            )
            warnings.append(msg)

        high_cardinality_features = self.find_high_cardinality(df[features])
        if features_for_generate:
            high_cardinality_features = [
                f for f in high_cardinality_features if columns_renaming.get(f, f) not in features_for_generate
            ]
        if high_cardinality_features:
            msg = bundle.get("high_cardinality_features").format(
                [columns_renaming.get(f, f) for f in high_cardinality_features]
            )
            warnings.append(msg)

        return (empty_or_constant_features + high_cardinality_features, warnings)

    @staticmethod
    def find_high_cardinality(df: pd.DataFrame) -> List[str]:
        # Remove high cardinality columns
        row_count = df.shape[0]
        if row_count < 100:  # For tests with small datasets
            return []
        return [
            i
            for i in df
            if (is_object_dtype(df[i]) or is_string_dtype(df[i]) or FeaturesValidator.__is_integer(df[i]))
            and (df[i].nunique(dropna=False) / row_count >= 0.85)
        ]

    @staticmethod
    def __is_integer(series: pd.Series) -> bool:
        return (
            is_integer_dtype(series)
            or series.dropna()
            .apply(
                lambda f: (float.is_integer(f) and abs(f) < np.iinfo(np.int64).max) if isinstance(f, float) else False
            )
            .all()
        )

    @staticmethod
    def find_constant_features(df: pd.DataFrame) -> List[str]:
        return [i for i in df if df[i].nunique() <= 1]

    @staticmethod
    def is_one_hot_encoded(series: pd.Series) -> bool:
        try:
            # All rows should be the same type
            if series.apply(lambda x: type(x)).nunique() != 1:
                return False

            # First, handle string representations of True/False
            series_copy = series.copy()
            if series_copy.dtype == "object" or series_copy.dtype == "string":
                # Convert string representations of boolean values to numeric
                series_copy = series_copy.astype(str).str.strip().str.lower()
                series_copy = series_copy.replace({"true": "1", "false": "0"})

            # Column contains only 0 and 1 (as strings or numbers or booleans)
            series_copy = series_copy.astype(float)
            if set(series_copy.unique()) != {0.0, 1.0}:
                return False

            series_copy = series_copy.astype(int)

            # Column doesn't contain any NaN, np.NaN, space, null, etc.
            if not (series_copy.isin([0, 1])).all():
                return False

            vc = series_copy.value_counts()
            # Column should contain both 0 and 1
            if len(vc) != 2:
                return False

            # Minority class is 1
            if vc[1] >= vc[0]:
                return False

            return True
        except ValueError:
            return False
